@@ -1,8 +1,6 @@
-BEGIN { $INC{$_} = 1 for qw(Parse/HTTP/UserAgent.pm Parse/HTTP/UserAgent/Constants.pm Parse/HTTP/UserAgent/Base/Accessors.pm Parse/HTTP/UserAgent/Base/Dumper.pm Parse/HTTP/UserAgent/Base/IS.pm Parse/HTTP/UserAgent/Base/Parsers.pm); }
+BEGIN { $INC{$_} = 1 for qw(Parse/HTTP/UserAgent.pm Parse/HTTP/UserAgent/Base/Accessors.pm Parse/HTTP/UserAgent/Base/Dumper.pm Parse/HTTP/UserAgent/Base/IS.pm Parse/HTTP/UserAgent/Base/Parsers.pm Parse/HTTP/UserAgent/Constants.pm); }
 package Parse::HTTP::UserAgent;
 sub ________monolith {}
-package Parse::HTTP::UserAgent::Constants;
-sub ________monolith {}
 package Parse::HTTP::UserAgent::Base::Accessors;
 sub ________monolith {}
 package Parse::HTTP::UserAgent::Base::Dumper;
@@ -12,19 +10,22 @@ sub ________monolith {}
 package Parse::HTTP::UserAgent::Base::Parsers;
 sub ________monolith {}
 package Parse::HTTP::UserAgent::Constants;
+sub ________monolith {}
+package Parse::HTTP::UserAgent::Constants;
 use strict;
 use warnings;
 use vars qw( $VERSION $OID @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS );
 
-$VERSION = '0.20';
+$VERSION = '0.35';
 
-use constant MINUS_ONE           => -1;
+use constant INIT_FIELD_COUNTER  => -1;
 use constant NO_IMATCH           => -1; # for index()
 use constant LAST_ELEMENT        => -1;
 
-BEGIN { $OID = MINUS_ONE }
+BEGIN { $OID = INIT_FIELD_COUNTER }
 
 use constant UA_STRING           => ++$OID; # just for information
+use constant UA_STRING_ORIGINAL  => ++$OID; # just for information
 use constant UA_UNKNOWN          => ++$OID; # failed to detect?
 use constant UA_GENERIC          => ++$OID; # parsed with a generic parser.
 use constant UA_NAME             => ++$OID; # The identifier of the ua
@@ -39,7 +40,8 @@ use constant UA_STRENGTH         => ++$OID; # [MSIE] List of .NET CLR versions
 use constant UA_MOZILLA          => ++$OID; # [Firefox] Mozilla revision
 use constant UA_ROBOT            => ++$OID; # Is this a robot?
 use constant UA_WAP              => ++$OID; # unimplemented
-use constant UA_MOBILE           => ++$OID; # unimplemented
+use constant UA_MOBILE           => ++$OID; # partially implemented
+use constant UA_TABLET           => ++$OID; # partially implemented
 use constant UA_PARSER           => ++$OID; # the parser name
 use constant UA_DEVICE           => ++$OID; # the name of the mobile device
 use constant UA_ORIGINAL_NAME    => ++$OID; # original name if this is some variation
@@ -53,11 +55,13 @@ use constant TK_NAME             => 0;
 use constant TK_ORIGINAL_VERSION => 1;
 use constant TK_VERSION          => 2;
 
+use constant INSIDE_UNIT_TEST    => $ENV{PARSE_HTTP_USERAGENT_TEST_SUITE};
+use constant INSIDE_VERBOSE_TEST => INSIDE_UNIT_TEST && $ENV{HARNESS_IS_VERBOSE};
 use constant RE_FIREFOX_NAMES    => qr{Firefox|Iceweasel|Firebird|Phoenix }xms;
-use constant RE_DOTNET           => qr{ \A [.]NET \s+ CLR \s+ (.+?) \z    }xms;
+use constant RE_DOTNET           => qr{ \A [.]NET (?: \s+ CLR \s+ )? (.+?) \z    }xms;
 use constant RE_WINDOWS_OS       => qr{ \A Win(dows|NT|[0-9]+)?           }xmsi;
 use constant RE_SLASH            => qr{ /                                 }xms;
-use constant RE_SPLIT_PARSE      => qr{ \s? [()] \s?                      }xms;
+use constant RE_SPLIT_PARSE      => qr{ \s? ([()]) \s?                    }xms;
 use constant RE_OPERA_MINI       => qr{ \A (Opera \s+ Mini) / (.+?) \z    }xms;
 use constant RE_TRIDENT          => qr{ \A (Trident) / (.+?) \z           }xmsi;
 use constant RE_EPIPHANY_GECKO   => qr{ \A (Epiphany) / (.+?) \z          }xmsi;
@@ -74,7 +78,13 @@ use constant RE_TWO_LETTER_LANG  => qr{ \A [a-z]{2} \z }xms;
 use constant RE_DIGIT_DOT_DIGIT  => qr{\d+[.]?\d}xms;
 
 use constant RE_WARN_OVERFLOW => qr{\QInteger overflow in version\E}xms;
-use constant RE_WARN_INVALID  => qr{\QVersion string .+? contains invalid data; ignoring:\E}xms;
+use constant RE_WARN_INVALID  => qr{\QVersion string\E .+? \Qcontains invalid data; ignoring:\E}xms;
+
+use constant ERROR_MAXTHON_VERSION  => 'Unable to extract Maxthon version from Maxthon UA-string';
+use constant ERROR_MAXTHON_MSIE     => 'Unable to extract MSIE from Maxthon UA-string';
+use constant OPERA9                 => 9;
+use constant OPERA_TK_LENGTH        => 5;
+use constant OPERA_FAKER_EXTRA_SIZE => 4;
 
 use constant LIST_ROBOTS         => qw(
     Wget
@@ -84,6 +94,7 @@ use constant LIST_ROBOTS         => qw(
     Googlebot
     Baiduspider+
     msnbot
+    bingbot
 ), 'Yahoo! Slurp';
 
 use base qw( Exporter );
@@ -95,6 +106,7 @@ BEGIN {
             IS_MAXTHON
             IS_EXTENDED
             UA_STRING
+            UA_STRING_ORIGINAL
             UA_UNKNOWN
             UA_GENERIC
             UA_NAME
@@ -110,6 +122,7 @@ BEGIN {
             UA_ROBOT
             UA_WAP
             UA_MOBILE
+            UA_TABLET
             UA_PARSER
             UA_DEVICE
             UA_ORIGINAL_NAME
@@ -150,6 +163,17 @@ BEGIN {
         etc => [qw(
             NO_IMATCH
             LAST_ELEMENT
+            INSIDE_UNIT_TEST
+            INSIDE_VERBOSE_TEST
+        )],
+        error => [qw(
+            ERROR_MAXTHON_VERSION
+            ERROR_MAXTHON_MSIE
+        )],
+        opera => [qw(
+            OPERA9
+            OPERA_TK_LENGTH
+            OPERA_FAKER_EXTRA_SIZE
         )],
     );
 
@@ -162,12 +186,8 @@ use strict;
 use warnings;
 use vars qw( $VERSION );
 use Parse::HTTP::UserAgent::Constants qw(:all);
-use constant ERROR_MAXTHON_VERSION => 'Unable to extract Maxthon version from Maxthon UA-string';
-use constant ERROR_MAXTHON_MSIE    => 'Unable to extract MSIE from Maxthon UA-string';
-use constant OPERA9                => 9;
-use constant OPERA_TK_LENGTH       => 5;
 
-$VERSION = '0.20';
+$VERSION = '0.35';
 
 sub _extract_dotnet {
     my($self, @args) = @_;
@@ -179,9 +199,12 @@ sub _extract_dotnet {
             push @dotnet, $match[0];
             next;
         }
-        if ( $e =~ RE_WINDOWS_OS && $1 ne '64' ) {
-            $self->[UA_OS] = $e;
-            next;
+        if ( $e =~ RE_WINDOWS_OS ) {
+            if ( $1 && $1 ne '64' ) {
+                # Maxthon stupidity: multiple OS definitions
+                $self->[UA_OS] ||= $e;
+                next;
+            }
         }
         push @extras, $e;
     }
@@ -227,9 +250,17 @@ sub _parse_maxthon {
     my($self, $moz, $thing, $extra, @others) = @_;
     my @omap = grep { $_ } map { split RE_SC_WS_MULTI, $_ } @others;
     my($maxthon, $msie, @buf);
+
     foreach my $e ( @omap, @{$thing} ) { # $extra -> junk
-        if ( index(uc $e, 'MAXTHON') != NO_IMATCH ) { $maxthon = $e; next; }
-        if ( index(uc $e, 'MSIE'   ) != NO_IMATCH ) { $msie    = $e; next; }
+        if ( index(uc $e, 'MAXTHON') != NO_IMATCH ) {
+            $maxthon = $e;
+            next;
+        }
+        if ( index(uc $e, 'MSIE' ) != NO_IMATCH ) {
+            # Maxthon stupidity: multiple MSIE strings
+            $msie ||= $e;
+            next;
+        }
         push @buf, $e;
     }
 
@@ -257,6 +288,7 @@ sub _parse_maxthon {
 
     $self->[UA_ORIGINAL_VERSION] = $v;
     $self->[UA_ORIGINAL_NAME]    = 'Maxthon';
+    $self->[UA_PARSER]           = 'maxthon';
     return 1;
 }
 
@@ -285,36 +317,120 @@ sub _parse_msie {
         }
         push @buf, $e;
     }
-    $self->[UA_EXTRAS] = [ @buf ];
+
+    $self->[UA_EXTRAS] = [
+        map  { $self->trim( $_ ) }
+        grep { $_ !~ m{ \s+ compatible \z }xms }
+        @buf
+    ];
+
     $self->[UA_PARSER] = 'msie';
+
     return 1;
 }
 
 sub _parse_firefox {
     my($self, @args) = @_;
     $self->_parse_mozilla_family( @args );
+    my $e = $self->[UA_EXTRAS];
+
+    if ( ref $e eq 'ARRAY'
+        && @{ $e } > 0
+        && index( lc $e->[-1], 'fennec' ) != NO_IMATCH
+    ) {
+        $self->_fix_fennec( $e );
+    }
+
     $self->[UA_NAME] = 'Firefox';
+
     return 1;
 }
 
+sub _fix_fennec {
+    my($self, $e) = @_;
+    my($name, $version) = split RE_SLASH, pop @{ $e };
+    $self->[UA_ORIGINAL_NAME]    = $name;
+    $self->[UA_ORIGINAL_VERSION] = $version;
+    $self->[UA_MOBILE]           = 1;
+    return if ! $self->[UA_LANG];
+
+    if ( lc $self->[UA_LANG] eq 'tablet' ) {
+        $self->[UA_TABLET] = 1;
+        $self->[UA_LANG]   = undef;
+    }
+    elsif ( index( $self->[UA_LANG], q{ } ) != NO_IMATCH ) {
+        push @{ $self->[UA_EXTRAS] }, $self->[UA_LANG];
+        $self->[UA_LANG] = undef;
+    }
+    else {
+        # Do nothing
+    }
+
+    return;
+}
+
 sub _parse_safari {
     my($self, $moz, $thing, $extra, @others) = @_;
-    my($version, @junk)     = split RE_WHITESPACE, pop @others;
-    my $ep = $version && index( lc($version), 'epiphany' ) != NO_IMATCH;
-    (undef, $version)       = split RE_SLASH, $version;
-    $self->[UA_NAME]        = $ep ? 'Epiphany' : 'Safari';
-    $self->[UA_VERSION_RAW] = $version;
+    my $ipad            = $thing && lc( $thing->[0] || q{} ) eq 'ipad';
+    my($version, @junk) = split RE_WHITESPACE, pop @others;
+    my $ep              = $version &&
+                            index( lc($version), 'epiphany' ) != NO_IMATCH;
+    my($junkv, $vx)     = split RE_SLASH, $version;
+
+    if ( $ipad ) {
+        shift @{ $thing }; # remove iPad
+        if ( $junkv && $junkv eq 'Mobile' ) {
+            unshift @junk, join q{/}, $junkv, $vx;
+            $vx = undef;
+        }
+        $self->[UA_MOBILE] = 1;
+        $self->[UA_TABLET] = 1;
+    }
+
+    $self->[UA_NAME]        = $ep   ? 'Epiphany'
+                            : $ipad ? 'iPad'
+                            :         'Safari';
+    $self->[UA_VERSION_RAW] = $vx;
     $self->[UA_TOOLKIT]     = $extra ? [ split RE_SLASH, $extra->[0] ] : [];
-    $self->[UA_LANG]        = pop @{ $thing };
+    if ( $thing->[-1] && length($thing->[LAST_ELEMENT]) <= 5 ) {
+        # todo: $self->_is_lang_field($junk)
+        # in here or in _post_parse()
+        $self->[UA_LANG]    = pop @{ $thing };
+    }
     $self->[UA_OS]          = @{$thing} && length $thing->[LAST_ELEMENT] > 1
                             ? pop   @{ $thing }
                             : shift @{ $thing }
                             ;
-    $self->[UA_DEVICE]      = shift @{$thing} if $thing->[0] eq 'iPhone';
-    $self->[UA_EXTRAS]      = [ @{$thing}, @others ];
+    if ( $self->[UA_OS] && lc $self->[UA_OS] eq 'macintosh' ) {
+        $self->[UA_OS]   = $self->[UA_LANG];
+        $self->[UA_LANG] = undef;
+    }
 
-    if ( length($self->[UA_OS]) == 1 ) {
-        push @{$self->[UA_EXTRAS]}, $self->[UA_EXTRAS];
+    if ( $thing->[0] && lc $thing->[0] eq 'iphone' ) {
+        $self->[UA_MOBILE] = 1;
+        $self->[UA_DEVICE] = shift @{$thing};
+        my $check_os       = $thing->[LAST_ELEMENT];
+
+        if ( $check_os && index( $check_os, 'Mac OS X' ) != NO_IMATCH ) {
+            if ( $self->[UA_OS] ) {
+                push @{$self->[UA_EXTRAS]}, $self->[UA_OS];
+            }
+            $self->[UA_OS] = pop @{ $thing };
+            # Another oddity: tk as "AppleWebKit/en_SG"
+            if ( ! $self->[UA_LANG] && $self->[UA_TOOLKIT] ) {
+                my $v = $self->[UA_TOOLKIT][TK_ORIGINAL_VERSION];
+                if ( $v && $v =~ m< [a-zA-Z]{2}_[a-zA-Z]{2} >xms ) {
+                    $self->[UA_LANG] = $v;
+                    $self->[UA_TOOLKIT][TK_ORIGINAL_VERSION] = undef;
+                }
+            }
+        }
+    }
+
+    $self->[UA_EXTRAS] = [ @{$thing}, @others ];
+
+    if ( $self->[UA_OS] && length($self->[UA_OS]) == 1 ) {
+        push @{$self->[UA_EXTRAS]}, $self->[UA_OS];
         $self->[UA_OS] = undef;
     }
 
@@ -335,6 +451,66 @@ sub _parse_chrome {
     return 1;
 }
 
+sub _parse_android {
+    my($self, $moz, $thing, $extra, @others) = @_;
+    (undef, @{$self}[UA_STRENGTH, UA_OS, UA_LANG, UA_DEVICE]) = @{ $thing };
+    if ( ! $extra
+        && $others[0]
+        && index( $others[0], 'AppleWebKit' ) != NO_IMATCH
+    ) {
+        $extra = [ shift @others ];
+        $self->[UA_PARSER] = 'android:paren_fixer';
+    }
+    $self->[UA_TOOLKIT] = [ split RE_SLASH, $extra->[0] ] if $extra;
+    my(@extras, $is_phone);
+
+    my @junkions = map { split m{\s+}xms } @others;
+    foreach my $junk ( @junkions ) {
+        if ( $junk eq 'Mobile' ) {
+            $is_phone = 1;
+            next;
+        }
+        if ( index( $junk, 'Version' ) != NO_IMATCH ) {
+            my(undef, $v) = split RE_SLASH, $junk;
+            $self->[UA_VERSION_RAW] = $v; # looks_like_number?
+            next;
+        }
+        push @extras, $junk;
+    }
+
+    if ( $self->[UA_DEVICE] ) {
+        my @build = split RE_WHITESPACE, $self->[UA_DEVICE];
+        my @btest;
+        while ( @build && index($build[-1], 'Build') == NO_IMATCH ) {
+            unshift @btest, pop @build;
+        }
+        unshift @btest, pop @build if @build;
+        my $device = @build ? join ' ', @build : undef;
+        my $build  = shift @btest;
+
+        if ( $device && $build ) {
+            $build =~ s{ Build/ }{}xms;
+            my $os = $self->[UA_OS] || 'Android';
+            $self->[UA_DEVICE] = $device;
+            $self->[UA_OS]     = "$os ($build)";
+            if ( @btest ) {
+                $self->[UA_TOOLKIT] = [ split RE_SLASH, $btest[0] ];
+            }
+        }
+    }
+
+    if ( @extras >= 3 && $extras[0] && $extras[0] eq 'KHTML,') {
+        unshift @extras, join ' ', map { shift @extras } 1..3;
+    }
+
+    $self->[UA_NAME]   = 'Android';
+    $self->[UA_MOBILE] = 1;
+    $self->[UA_TABLET] = $is_phone ? undef : 1;
+    $self->[UA_EXTRAS] = [ grep { $_ } @extras ];
+
+    return 1;
+}
+
 sub _parse_opera_pre {
     # opera 5,9
     my($self, $moz, $thing, $extra) = @_;
@@ -397,22 +573,42 @@ sub _parse_opera_post {
 }
 
 sub _parse_mozilla_family {
-    my($self, $moz, $thing, $extra, @extras) = @_;
+    my($self, $moz, $thing, $extra, @others) = @_;
     # firefox variation or just mozilla itself
     my($name, $version)      = split RE_SLASH, defined $extra->[1] ? $extra->[1]
                              :                                       $moz
                              ;
+    if ( $version ) {
+        $extra->[1] = '';
+    }
     $self->[UA_NAME]         = $name;
-    $self->[UA_TOOLKIT]      = $extra ? [ split RE_SLASH, $extra->[0] ] : [];
     $self->[UA_VERSION_RAW]  = $version;
+    $self->[UA_TOOLKIT]      = $extra->[0]
+                             ? [ split RE_SLASH, shift @{ $extra } ]
+                             : []
+                             ;
 
     if ( @{$thing} && index($thing->[LAST_ELEMENT], 'rv:') != NO_IMATCH ) {
         $self->[UA_MOZILLA]  = pop @{ $thing };
-        $self->[UA_LANG]     = pop @{ $thing };
-        $self->[UA_OS]       = pop @{ $thing };
+        if ( @{ $thing } <= 3 ) {
+            $self->[UA_OS] = shift @{ $thing };
+            if ( $self->[UA_OS] && $self->[UA_OS] eq 'Macintosh' ) {
+                $self->[UA_OS] = shift @{ $thing };
+            }
+            $self->[UA_LANG] = pop @{ $thing } if @{ $thing };
+        }
+        else {
+            $self->[UA_LANG]     = pop @{ $thing };
+            $self->[UA_OS]       = pop @{ $thing };
+        }
     }
 
-    $self->[UA_EXTRAS] = [ @{ $thing }, @extras ];
+    $self->[UA_EXTRAS] = [
+        grep { $_ }
+        @{ $thing },
+        @others,
+        $extra ? @{ $extra } : (),
+    ];
     return 1;
 }
 
@@ -623,6 +819,7 @@ sub _parse_emacs {
     $self->[UA_NAME]        = $name;
     $self->[UA_VERSION_RAW] = $version || 0;
     $self->[UA_OS]          = shift @{ $thing };
+    $self->[UA_OS]          = $self->trim( $self->[UA_OS] ) if $self->[UA_OS];
     my @rest = (  @{ $thing }, @moz );
     push @rest, @{ $extra } if $extra && ref $extra eq 'ARRAY';
     push @rest, ( map { split RE_SC_WS, $_ } @others ) if @others;
@@ -687,9 +884,8 @@ use strict;
 use warnings;
 use vars qw( $VERSION );
 use Parse::HTTP::UserAgent::Constants qw(:all);
-use constant OPERA_FAKER_EXTRA_SIZE => 4;
 
-$VERSION = '0.20';
+$VERSION = '0.35';
 
 sub _is_opera_pre {
     my($self, $moz) = @_;
@@ -703,17 +899,20 @@ sub _is_opera_post {
 
 sub _is_opera_ff { # opera faking as firefox
     my($self, $extra) = @_;
-    return $extra && @{$extra} == OPERA_FAKER_EXTRA_SIZE && $extra->[2] eq 'Opera';
+    return $extra
+            && @{$extra}    ==  OPERA_FAKER_EXTRA_SIZE
+            &&  $extra->[2] eq 'Opera';
 }
 
 sub _is_safari {
     my($self, $extra, $others) = @_;
     my $str = $self->[UA_STRING];
     # epiphany?
-    return                index( $str                   , 'Chrome'       ) != NO_IMATCH ? 0 # faker
-          :    $extra  && index( $extra->[0]            , 'AppleWebKit'  ) != NO_IMATCH ? 1
-          : @{$others} && index( $others->[LAST_ELEMENT], 'Safari'       ) != NO_IMATCH ? 1
-          :                                                                     0
+    return                index( $str                   , 'Chrome'      ) != NO_IMATCH ? 0 # faker
+          :               index( $str                   , 'Android'     ) != NO_IMATCH ? 0 # faker
+          :    $extra  && index( $extra->[0]            , 'AppleWebKit' ) != NO_IMATCH ? 1
+          : @{$others} && index( $others->[LAST_ELEMENT], 'Safari'      ) != NO_IMATCH ? 1
+          :                                                                              0
           ;
 }
 
@@ -728,6 +927,29 @@ sub _is_chrome {
            ( $extra  && index( $extra->[0], 'AppleWebKit') != NO_IMATCH);
 }
 
+sub _is_android {
+    my($self, $thing, $others) = @_;
+    my $has_android = grep { index( lc $_, 'android' ) != NO_IMATCH  } @{ $thing  };
+    my $has_safari  = grep { index( lc $_, 'safari'  ) != NO_IMATCH  } @{ $others };
+    if ( $has_android && $has_safari ) {
+        return 1;
+    }
+    if (   @{ $others } == 0
+        && @{ $thing  }  > 0
+        && $thing->[-1]
+        && index( $thing->[-1], 'AppleWebKit' ) != NO_IMATCH
+    ) {
+        # More stupidity: ua string is missing a closing paren
+        my($part, @rest) = split m{(AppleWebKit)}xms, $thing->[-1];
+        $thing->[-1] = $part;
+        @{ $others } =  map   { $self->trim( $_ ) }
+                        split m{ (\QKHTML, like Gecko\E) }xms,
+                        join  q{}, @rest;
+        return 1;
+    }
+    return;
+}
+
 sub _is_ff {
     my($self, $extra) = @_;
     return if ! $extra || ! $extra->[1];
@@ -806,14 +1028,14 @@ package Parse::HTTP::UserAgent::Base::Dumper;
 use strict;
 use warnings;
 use vars qw( $VERSION );
-use Parse::HTTP::UserAgent::Constants qw(:all);
 use Carp qw( croak );
+use Parse::HTTP::UserAgent::Constants qw(:all);
 
-$VERSION = '0.20';
+$VERSION = '0.35';
 
 sub dumper {
     my($self, @args) = @_;
-    my %opt  = @args % 2 ? () : (
+    my %opt = @args % 2 ? () : (
         type      => 'dumper',
         format    => 'none',
         interpret => 0,
@@ -877,6 +1099,7 @@ sub _dumper_dumper {
                         $titles[1],
                         q{-} x $max, q{ } x 2, q{-} x ($max*2);
     require Data::Dumper;
+    my @buf;
     foreach my $id ( @ids ) {
         my $name = $args ? $id->{name} : $id;
         my $val  = $args ? $id->{value} : $self->[ $self->$id() ];
@@ -888,11 +1111,14 @@ sub _dumper_dumper {
                     $rv =~ s{ ; }{}xms;
                     $rv eq '[]' ? q{} : $rv;
                 } if $val && ref $val;
-        $buf .= sprintf "%s%s%s\n",
+        push @buf, [
                         $name,
                         (q{ } x (2 + $max - length $name)),
                         defined $val ? $val : q{}
-                        ;
+                    ];
+    }
+    foreach my $row ( sort { lc $a->[0] cmp lc $b->[0] } @buf ) {
+        $buf .= sprintf "%s%s%s\n", @{ $row };
     }
     return $buf;
 }
@@ -903,23 +1129,52 @@ use warnings;
 use vars qw( $VERSION );
 use Parse::HTTP::UserAgent::Constants qw(:all);
 
-$VERSION = '0.20';
+$VERSION = '0.35';
 
 #TODO: new accessors
 #wap
 #mobile
 #device
+#tablet
 
-sub name             { return shift->[UA_NAME]             || q{} }
-sub unknown          { return shift->[UA_UNKNOWN]          || q{} }
-sub generic          { return shift->[UA_GENERIC]          || q{} }
-sub os               { return shift->[UA_OS]               || q{} }
-sub lang             { return shift->[UA_LANG]             || q{} }
-sub strength         { return shift->[UA_STRENGTH]         || q{} }
-sub parser           { return shift->[UA_PARSER]           || q{} }
-sub original_name    { return shift->[UA_ORIGINAL_NAME]    || q{} }
-sub original_version { return shift->[UA_ORIGINAL_VERSION] || q{} }
-sub robot            { return shift->[UA_ROBOT]            ||   0 }
+BEGIN {
+    my @simple = qw(
+        name
+        unknown
+        generic
+        os
+        lang
+        strength
+        parser
+        original_name
+        original_version
+        robot
+    );
+
+    my @multi = qw(
+        mozilla
+        extras
+        dotnet
+    );
+
+    no strict qw(refs); ## no critic (TestingAndDebugging::ProhibitNoStrict)
+    foreach my $name ( @simple ) {
+        my $id = 'UA_' . uc $name;
+        $id = __PACKAGE__->$id();
+        *{ $name } = sub { return shift->[$id] || q{} };
+    }
+
+    foreach my $name ( @multi ) {
+        my $id = 'UA_' . uc $name;
+        $id = __PACKAGE__->$id();
+        *{ $name } = sub {
+            my $self = shift;
+            return +() if ! $self->[ $id ];
+            my @rv = @{ $self->[ $id ] };
+            return wantarray ? @rv : $rv[0];
+        };
+    }
+}
 
 sub version {
     my $self = shift;
@@ -927,29 +1182,37 @@ sub version {
     return $self->[ $type eq 'raw' ? UA_VERSION_RAW : UA_VERSION ] || 0;
 }
 
-sub mozilla {
-    my $self = shift;
-    return +() if ! $self->[UA_MOZILLA];
-    my @rv = @{ $self->[UA_MOZILLA] };
-    return wantarray ? @rv : $rv[0];
-}
-
 sub toolkit {
     my $self = shift;
-    return +() if ! $self->[UA_TOOLKIT];
-    return @{ $self->[UA_TOOLKIT] };
+    return Parse::HTTP::UserAgent::Base::Accessors::toolkit->new(
+                $self->[UA_TOOLKIT]
+            );
 }
 
-sub extras {
-    my $self = shift;
-    return +() if ! $self->[UA_EXTRAS];
-    return @{ $self->[UA_EXTRAS] };
+package Parse::HTTP::UserAgent::Base::Accessors::toolkit;
+use strict;
+use warnings;
+use overload '""',    => 'name',
+             '0+',    => 'version',
+             fallback => 1,
+;
+use constant ID_NAME        => 0;
+use constant ID_VERSION_RAW => 1;
+use constant ID_VERSION     => 2;
+
+sub new {
+    my($class, $tk) = @_;
+    return bless [ $tk ? @{ $tk } : (undef) x 3 ], $class;
 }
 
-sub dotnet {
+sub name {
+    return shift->[ID_NAME];
+}
+
+sub version {
     my $self = shift;
-    return +() if ! $self->[UA_DOTNET];
-    return @{ $self->[UA_DOTNET] };
+    my $type = shift || q{};
+    return $self->[ $type eq 'raw' ? ID_VERSION_RAW : ID_VERSION ] || 0;
 }
 
 package Parse::HTTP::UserAgent;
@@ -957,7 +1220,7 @@ use strict;
 use warnings;
 use vars qw( $VERSION );
 
-$VERSION = '0.20';
+$VERSION = '0.35';
 
 use base qw(
     Parse::HTTP::UserAgent::Base::IS
@@ -999,8 +1262,9 @@ sub new {
     croak 'Options must be a hash reference' if ref $opt ne 'HASH';
     my $self  = [ map { undef } 0..MAXID ];
     bless $self, $class;
-    $self->[UA_STRING]   = $ua;
+    @{ $self }[ UA_STRING, UA_STRING_ORIGINAL ] = ($ua) x 2;
     $self->[IS_EXTENDED] = exists $opt->{extended} ? $opt->{extended} : 1;
+    $self->_normalize( $opt->{normalize} ) if $opt->{normalize};
     $self->_parse;
     return $self;
 }
@@ -1024,6 +1288,25 @@ sub trim {
     return $s;
 }
 
+sub _normalize {
+    my $self = shift;
+    my $nopt = shift;
+    my $type = ref $nopt;
+
+    my @o = ! $type            ? ':all'
+          :   $type eq 'ARRAY' ? @{ $nopt }
+          :                      croak "Normalization option $nopt is invalid";
+
+    my %mode      = map { $_ => 1 } @o;
+    my @all       = qw( plus_to_space trim_spaces );
+    @mode{ @all } = (1) x @all if delete $mode{':all'};
+
+    my $s = \$self->[UA_STRING];
+    ${$s} =~ s{[+]}{ }xmsg if $mode{plus_to_space};
+    ${$s} =~ s<\s+>< >xmsg if $mode{trim_spaces};
+    return;
+}
+
 sub _parse {
     my $self = shift;
     return $self if $self->[IS_PARSED];
@@ -1037,9 +1320,32 @@ sub _pre_parse {
     my $self = shift;
     $self->[IS_MAXTHON] = index(uc $self->[UA_STRING], 'MAXTHON') != NO_IMATCH;
     my $ua = $self->[UA_STRING];
-    my($moz, $thing, $extra, @others) = split RE_SPLIT_PARSE, $ua;
+
+    my @parts;
+    my $i     = 0;
+    my $depth = 0;
+    foreach my $token ( split RE_SPLIT_PARSE, $ua ) {
+        if ( $token eq '(' ) {
+            $i++ if ++$depth == 1;
+            next;
+        }
+        if ( $token eq ')' ) {
+            $i++ if --$depth == 0;
+            next;
+        }
+        push @{ $parts[$i] ||= [] }, $token;
+    }
+
+    # Hopefully the above code was successful and now we can set the actual
+    # tokens to use inside parsers.
+    my($moz)    = join ' ', @{ shift(@parts) || []  };
+    my($thing)  = join ' ', @{ shift(@parts) || []  };
+    my($extra)  = join ' ', @{ shift(@parts) || []  };
+    my(@others) = map { @{ $_ } } @parts;
+
     $thing = $thing ? [ split RE_SC_WS, $thing ] : [];
     $extra = [ split RE_WHITESPACE, $extra ] if $extra;
+
     $self->_debug_pre_parse( $moz, $thing, $extra, @others ) if DEBUG;
     return $moz, $thing, $extra, @others;
 }
@@ -1055,13 +1361,14 @@ sub _do_parse {
         }
     }
 
-    my $rv =  $self->_is_opera_pre($m)   ? [opera_pre  => $m, $t, $e           ]
+    my $rv =  $self->[IS_MAXTHON]        ? [maxthon    => $m, $t, $e, @o       ]
+            : $self->_is_opera_pre($m)   ? [opera_pre  => $m, $t, $e           ]
             : $self->_is_opera_post($e)  ? [opera_post => $m, $t, $e, $c       ]
             : $self->_is_opera_ff($e)    ? [opera_pre  => "$e->[2]/$e->[3]", $t]
             : $self->_is_ff($e)          ? [firefox    => $m, $t, $e, @o       ]
             : $self->_is_safari($e, \@o) ? [safari     => $m, $t, $e, @o       ]
             : $self->_is_chrome($e, \@o) ? [chrome     => $m, $t, $e, @o       ]
-            : $self->[IS_MAXTHON]        ? [maxthon    => $m, $t, $e, @o       ]
+            : $self->_is_android($t,\@o) ? [android    => $m, $t, $e, @o       ]
             : undef;
 
     if ( $rv ) {
@@ -1069,7 +1376,7 @@ sub _do_parse {
         my $method = '_parse_' . $pname;
         my $rvx    = $self->$method( @{ $rv } );
         if ( $rvx ) {
-            $self->[UA_PARSER] = $pname;
+            $self->[UA_PARSER] ||= $pname;
             return $rvx;
         }
     }
@@ -1097,8 +1404,8 @@ sub _post_parse {
     $self->[UA_EXTRAS] = [ @buf ];
 
     if ( $self->[UA_TOOLKIT] ) {
-        push @{ $self->[UA_TOOLKIT] },
-             $self->_numify( $self->[UA_TOOLKIT][TK_ORIGINAL_VERSION] );
+        my $v = $self->[UA_TOOLKIT][TK_ORIGINAL_VERSION];
+        push @{ $self->[UA_TOOLKIT] }, defined $v ? $self->_numify( $v ) : 0;
     }
 
     if( $self->[UA_MOZILLA] ) {
@@ -1141,7 +1448,10 @@ sub _object_ids {
 sub _numify {
     my $self = shift;
     my $v    = shift || return 0;
-    $v    =~ s{
+    my @removed;
+
+    if (
+        $v =~ s{(
                 pre      |
                 rel      |
                 alpha    |
@@ -1150,8 +1460,46 @@ sub _numify {
                 gold     |
                 [ab]\d+  |
                 a\-XXXX  |
-                \+
-                }{}xmsig;
+                [+]
+               )}{}xmsig
+    ){
+        push @removed, $1 if INSIDE_VERBOSE_TEST;
+    }
+
+    if (
+        $v =~ s{(
+                (?:[^0-9]+)? # usually dash
+                rc           # nonsense
+                [\-_.]?      # usually dash
+                ([0-9])      # teh candidate revision
+            )}{.0.$2}xmsi    # yeah, hacky
+    ) {
+        push @removed, $1 if INSIDE_VERBOSE_TEST;
+    }
+
+    # workaround another stupidity (1.2.3-4)
+    if ( my $rc = $v =~ tr/-/./ ) {
+        push @removed, '-' x $rc if INSIDE_VERBOSE_TEST;
+    }
+
+    # Finally, be aggressive to prevent dying on bogus stuff.
+    # It's interesting how people provide highly stupid version "numbers".
+    # Version parameters are probably more stupid than the UA string itself.
+    if ( $v =~ s<([^0-9._v])><.>xmsg ) {
+        push @removed, $1 if INSIDE_VERBOSE_TEST;
+    }
+
+    if ( $v =~ s<([.]{2,})><.>xmsg ) {
+        push @removed, $1 if INSIDE_VERBOSE_TEST;
+    }
+
+    if ( INSIDE_VERBOSE_TEST ) {
+        if ( @removed ) {
+            my $r = join q{','}, @removed;
+            Test::More::diag("[DEBUG] _numify: removed '$r' from version string");
+        }
+    }
+
     # Gecko revisions like: "20080915000512" will cause an
     #   integer overflow warning. use bigint?
     local $SIG{__WARN__} = sub {
@@ -1161,7 +1509,30 @@ sub _numify {
     # if version::vpp is used it'll identify 420 as a v-string
     # add a floating point to fool it
     $v .= q{.0} if index($v, q{.}) == NO_IMATCH;
-    my $rv = version->new("$v")->numify;
+    (my $check = $v) =~ tr/0-9//cd;
+    return 0 if ! $check; # A string parsed as version (i.e.: AppleWebKit/en_SG)
+    my $rv;
+    eval {
+        $rv = version->new("$v")->numify;
+        1;
+    } or do {
+        my $error = $@ || '[unknown error while parsing version]';
+        if ( INSIDE_UNIT_TEST ) {
+            chomp $error;
+            if ( INSIDE_VERBOSE_TEST ) {
+                Test::More::diag( "[FATAL] _numify: version said: $error for '$v'" );
+                Test::More::diag(
+                    sprintf '[FATAL] _numify: UA with bogus version (%s) is: %s',
+                                $v, $self->[UA_STRING]
+                );
+                Test::More::diag( '[FATAL] _numify: ' . $self->dumper );
+            }
+            croak $error;
+        }
+        else {
+            croak $error;
+        }
+    };
     return $rv;
 }
 
@@ -1208,8 +1579,8 @@ generated with an automatic build tool. If you experience problems
 with this version, please install and use the supported standard
 version. This version is B<NOT SUPPORTED>.
 
-This document describes version C<0.20> of C<Parse::HTTP::UserAgent>
-released on C<27 October 2009>.
+This document describes version C<0.35> of C<Parse::HTTP::UserAgent>
+released on C<14 May 2012>.
 
 Quoting L<http://www.webaim.org/blog/user-agent-string-history/>:
 
@@ -1227,8 +1598,8 @@ depending on the vendor's (or the user's) choice. Also, it is not dependable
 since it is some arbitrary identification string. Any user agent can fake
 another. So, why deal with such a useless mess? You may want to see the choice
 of your visitors and can get some reliable data (even if some are fake) and
-generate some nice charts out of them or just want to send a C<HttpOnly> cookie
-if the user agent seem to support it (and send a normal one if this is not the
+generate some nice charts out of them or just want to send an C<HttpOnly> cookie
+if the user agent seems to support it (and send a normal one if this is not the
 case). However, browser sniffing for client-side coding is considered a bad
 habit.
 
@@ -1315,11 +1686,11 @@ If you pass a wrong parameter to the dumper, it'll croak.
 
 =item *
 
-L<HTTP::BrowserDetect>
+L<HTML::ParseBrowser>
 
 =item *
 
-L<HTML::ParseBrowser>
+L<HTTP::BrowserDetect>
 
 =item *
 
@@ -1329,6 +1700,10 @@ L<HTTP::DetectUserAgent>
 
 L<HTTP::MobileAgent>
 
+=item *
+
+L<Mobile::UserAgent>
+
 =back
 
 =head2 Resources
@@ -1337,19 +1712,44 @@ L<HTTP::MobileAgent>
 
 =item *
 
-L<http://en.wikipedia.org/wiki/User_agent>,
+L<http://en.wikipedia.org/wiki/User_agent>
 
 =item *
 
-L<http://www.zytrax.com/tech/web/browser_ids.htm>,
+L<http://www.zytrax.com/tech/web/browser_ids.htm>
 
 =item *
 
-L<http://www.zytrax.com/tech/web/mobile_ids.html>,
+L<http://www.zytrax.com/tech/web/mobile_ids.html>
 
 =item *
 
-L<http://www.webaim.org/blog/user-agent-string-history/>.
+L<http://www.webaim.org/blog/user-agent-string-history/>
+
+=item *
+
+L<https://developer.mozilla.org/en/Gecko_user_agent_string_reference>
+
+=item *
+
+L<http://www.useragentstring.com>
+
+=back
+
+=head2 Module Reviews
+
+=over 4
+
+=item *
+
+CPAN modules for parsing User-Agent strings by B<Neil Bowers>:
+L<http://blogs.perl.org/users/neilb/2011/10/cpan-modules-for-parsing-user-agent-strings.html>
+(23 October 2011).
+
+=item *
+
+Parse::HTTP::UserAgent: yet another user agent string parser by B<Burak Gursoy>:
+L<http://use.perl.org/~Burak/journal/39577> (4 September 2009).
 
 =back
 
@@ -1359,12 +1759,12 @@ Burak Gursoy <burak@cpan.org>.
 
 =head1 COPYRIGHT
 
-Copyright 2009 Burak Gursoy. All rights reserved.
+Copyright 2009 - 2012 Burak Gursoy. All rights reserved.
 
 =head1 LICENSE
 
 This library is free software; you can redistribute it and/or modify 
-it under the same terms as Perl itself, either Perl version 5.10.0 or, 
+it under the same terms as Perl itself, either Perl version 5.12.3 or, 
 at your option, any later version of Perl 5 you may have available.
 
 =cut
